#include <stdint.h>
#include <stdbool.h>
#include "nrf.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "softdevice_handler.h"
#include "boards.h"
#include "nrf_log.h"

/**Constants for timeslot API
*/
static nrf_radio_request_t  m_timeslot_request;
static uint32_t             m_slot_length;

static volatile uint32_t total_timeslot_length;
static volatile uint32_t extend_success_count;
static volatile uint32_t fail_count;

static nrf_radio_signal_callback_return_param_t signal_callback_return_param;

#define TIMESLOT_EXTEND_LENGTH 200
#define TIMESLOT_TIMER_INTERRUPT_END_MARGIN		50		//margin in us on timer 0 interrupt before the timeslot ends

/**@brief Request next timeslot event in earliest configuration
 */
uint32_t request_next_event_earliest(void)
{
    m_slot_length                                  = TIMESLOT_EXTEND_LENGTH;
    m_timeslot_request.request_type                = NRF_RADIO_REQ_TYPE_EARLIEST;
    m_timeslot_request.params.earliest.hfclk       = NRF_RADIO_HFCLK_CFG_XTAL_GUARANTEED;
    m_timeslot_request.params.earliest.priority    = NRF_RADIO_PRIORITY_NORMAL;
    m_timeslot_request.params.earliest.length_us   = m_slot_length;
    m_timeslot_request.params.earliest.timeout_us  = 1000000;
    return sd_radio_request(&m_timeslot_request);
}

/**@brief Timeslot signal handler
 */
void nrf_evt_signal_handler(uint32_t evt_id)
{
    switch (evt_id)
    {
        case NRF_EVT_RADIO_SIGNAL_CALLBACK_INVALID_RETURN:
            NRF_LOG_INFO("NRF_EVT_RADIO_SIGNAL_CALLBACK_INVALID_RETURN\n");
            break;
        case NRF_EVT_RADIO_SESSION_IDLE:
			nrf_gpio_pin_toggle(LED_1);
			sd_radio_session_close();
            NRF_LOG_INFO("NRF_EVT_RADIO_SESSION_IDLE\n");
            break;
        case NRF_EVT_RADIO_SESSION_CLOSED:
            NRF_LOG_INFO("NRF_EVT_RADIO_SESSION_CLOSED\r\n");
			NRF_LOG_INFO(" extend_success_count            = %d\r\n", extend_success_count);
			NRF_LOG_INFO(" extend_fail_count               = %d\r\n", fail_count);
			NRF_LOG_INFO(" total timeslot length           = %d\r\n", total_timeslot_length);
            break;
        case NRF_EVT_RADIO_BLOCKED:
            NRF_LOG_INFO("NRF_EVT_RADIO_BLOCKED\n");
        case NRF_EVT_RADIO_CANCELED:
            NRF_LOG_INFO("NRF_EVT_RADIO_CANCELED\n");
            break;
        default:
            break;
    }
}

/**@brief Timeslot event handler
 */
nrf_radio_signal_callback_return_param_t * radio_callback(uint8_t signal_type)
{
    switch(signal_type)
    {
        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_START:
            
			NRF_LOG_INFO("NRF_RADIO_CALLBACK_SIGNAL_TYPE_START\n");

			NRF_TIMER0->INTENSET = TIMER_INTENSET_COMPARE0_Msk;
			NRF_TIMER0->BITMODE = (TIMER_BITMODE_BITMODE_32Bit << TIMER_BITMODE_BITMODE_Pos);
			NRF_TIMER0->PRESCALER = (4 << TIMER_PRESCALER_PRESCALER_Pos);
            NRF_TIMER0->CC[0] = m_slot_length - TIMESLOT_TIMER_INTERRUPT_END_MARGIN;
            NVIC_EnableIRQ(TIMER0_IRQn);
		
			total_timeslot_length = m_slot_length;
		
			signal_callback_return_param.params.request.p_next = NULL;
            signal_callback_return_param.params.extend.length_us = TIMESLOT_EXTEND_LENGTH;
            signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_EXTEND; 
            break;

        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_RADIO:
            signal_callback_return_param.params.request.p_next = NULL;
            signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_NONE;
            break;

        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_TIMER0:
			//timeslot is ending, signal the end action
            NRF_LOG_INFO("Timeslot ending after %d microseconds\n", NRF_TIMER0->CC[0]);
			signal_callback_return_param.params.request.p_next = NULL;
			signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_END;
            break;
        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_EXTEND_SUCCEEDED:
			
			extend_success_count++;
		
            if(total_timeslot_length >= 120000000)
			{
				signal_callback_return_param.params.request.p_next = NULL;
				signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_NONE;
			}
			else
			{
				total_timeslot_length += TIMESLOT_EXTEND_LENGTH;
				
				NRF_TIMER0->CC[0] = total_timeslot_length - TIMESLOT_TIMER_INTERRUPT_END_MARGIN;
				signal_callback_return_param.params.request.p_next = NULL;
				signal_callback_return_param.params.extend.length_us = TIMESLOT_EXTEND_LENGTH;
				signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_EXTEND; 
			}

			break;
        case NRF_RADIO_CALLBACK_SIGNAL_TYPE_EXTEND_FAILED:
            
            fail_count++;
			
			signal_callback_return_param.params.request.p_next = NULL;
            signal_callback_return_param.callback_action = NRF_RADIO_SIGNAL_CALLBACK_ACTION_END;
		
            break;
        default:
            //No implementation needed
            break;
    }
    return (&signal_callback_return_param);
}


/**@brief Function for initializing the timeslot API.
 */
uint32_t timeslot_sd_init(void)
{
    uint32_t err_code;
	
    err_code = sd_radio_session_open(radio_callback);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

    err_code = request_next_event_earliest();
    if (err_code != NRF_SUCCESS)
    {
        (void)sd_radio_session_close();
        return err_code;
    }
    return NRF_SUCCESS;
}
